<?php

namespace Drp\SimpleCms;

use Drp\SimpleCms\Contracts\Publishable;
use Drp\SimpleCms\Exceptions\NoUpdatesMade;

/**
 * Class ContentManager
 * @package Drp\SimpleCms
 */
class ContentManager implements \Drp\SimpleCms\Contracts\ContentManager
{
    /**
     * @var ContentFactory
     */
    protected $contentFactory;

    /**
     * ContentManager constructor.
     * @param ContentFactory $contentFactory
     */
    public function __construct(ContentFactory $contentFactory)
    {
        $this->contentFactory = $contentFactory;
    }

    /**
     * Create a new content of the given type
     *
     * @param string $contentType
     * @param mixed $attributes
     * @return Publishable
     */
    public function createContent($contentType, $attributes = null)
    {
        $content = $this->contentFactory->makeContentClass($contentType);
        $attributes['status'] = Publishable::STATUS_DRAFT;

        return $content->create($attributes);
    }

    /**
     * Processes updating a content item.
     *
     * Create a draft of the existing if it already exists
     * If the content is already in draft we will just update that version.
     *
     * @param Publishable $content
     * @param array|null $attributes The data to update
     * @return Publishable
     */
    public function updateContent(Publishable $content, $attributes = null)
    {
        $draft = $this->getDraft($content, $attributes);
        $callback = $content->beforeSave();

        if ($callback instanceof \Closure) {
            call_user_func_array($callback, [$draft, $attributes]);
        }

        $draft->fill($attributes);
        $draft->save();

        $callback = $content->afterSave();

        if ($callback instanceof \Closure) {
            call_user_func_array($callback, [$draft, $attributes]);
        }

        return $draft;
    }


    /**
     * Return the draft version of the given
     * content, If one does not exist create
     * it using the attributes.
     *
     * @param Publishable $content
     * @param array|string|null $attributes
     * @return Publishable Thrown if the Content model has not been updated.
     * @throws NoUpdatesMade Thrown if the Content model has not been updated.
     */
    protected function getDraft(Publishable $content, $attributes = null)
    {
        if (!$content->isContentDirty($attributes)) {
            throw new NoUpdatesMade();
        }

        $draft = $content->getDraftOrReviewVersion();

        if (empty($draft)) {
            //If a draft version does not exist create one
            return $this->createDraftCopy($content);
        }

        return $draft;
    }

    /**
     * Save a draft copy of the content with the new attributes.
     *
     * @param Publishable $content
     * @return Publishable
     */
    protected function createDraftCopy(Publishable $content)
    {
        $draft = $content->replicate();
        $draft->status = Publishable::STATUS_DRAFT;
        $draft->content_id = $content->content_id;
        $draft->save();

        return $draft;
    }
}
