<?php

namespace Drp\SimpleCms;

use Drp\SimpleCms\Contracts\Publishable;
use Drp\SimpleCms\Events\ContentWasPublished;
use Drp\SimpleCms\Exceptions\CouldNotResetDraft;
use Drp\SimpleCms\Exceptions\InvalidContentStatus;
use Illuminate\Database\Eloquent\Builder;

/**
 * Class Publisher
 * @package Drp\SimpleCms
 */
trait Publisher
{
    /**
     * Generate a unique Id for the content.
     * This is the next available id at the time of saving
     *
     * @return integer
     */
    public function generateContentId()
    {
        $last = $this->orderBy('content_id', true)->first();

        return is_null($last) ? 1 : $last->content_id + 1;
    }

    /**
     * Save the model to the database.
     *
     * @param  array $options
     * @return boolean
     */
    public function save(array $options = [])
    {
        $this->content_id = $this->content_id ?: $this->generateContentId();

        $this->ensureStatusIntegrity();

        return parent::save($options);
    }

    /**
     * Check if a published version of this content exists
     *
     * @throws InvalidContentStatus If there is a published version already.
     * @return void
     */
    private function ensureStatusIntegrity()
    {
        $existing = $this->versionAlreadyExists($this->id);

        if ($existing) {
            throw new InvalidContentStatus('Content already exists with this content_id and status');
        }
    }

    /**
     * Returns if the there is a version with the same status and contentId
     *
     * @param null|integer $currentId
     * @return boolean
     */
    private function versionAlreadyExists($currentId = null)
    {
        return $this
            ->where('content_id', $this->content_id)
            ->where('status', $this->status)
            ->where('id', '!=', $currentId)
            ->first();
    }

    /**
     * Return the published version of the current study.
     *
     * @throws ModelNotFoundException Thrown if no published version exists.
     * @return static
     */
    public function getPublishedVersion()
    {
        if ($this->isPublished()) {
            return $this;
        }

        return $this
            ->where('content_id', $this->content_id)
            ->wherePublished()
            ->first();
    }

    /**
     * Return the draft version of the current content.
     *
     * @throws \Illuminate\Database\Eloquent\ModelNotFoundException Thrown if no draft version exists.
     * @return static
     */
    public function getUnpublishedVersion()
    {
        if ($this->isUnPublished()) {
            return $this;
        }

        return $this
            ->where('content_id', $this->content_id)
            ->WhereUnPublished()
            ->first();
    }

    /**
     * Return the draft version of the current content.
     *
     * @throws \Illuminate\Database\Eloquent\ModelNotFoundException Thrown if no draft version exists.
     * @return static
     */
    public function getDraftOrReviewVersion()
    {
        if (!$this->isPublished() && !$this->isUnPublished()) {
            return $this;
        }

        return $this
            ->where('content_id', $this->content_id)
            ->whereDraftOrInReview()
            ->first();
    }

    /**
     * Scope a query to only include published content.
     *
     * @param \Illuminate\Database\Eloquent\Builder $query
     * @return \Illuminate\Database\Eloquent\Builder
     */
    public function scopeWherePublished(Builder $query)
    {
        return $query->where('status', Publishable::STATUS_PUBLISHED);
    }

    /**
     * Scope a query to only include un-published content.
     *
     * @param \Illuminate\Database\Eloquent\Builder $query
     * @return \Illuminate\Database\Eloquent\Builder
     */
    public function scopeWhereUnPublished(Builder $query)
    {
        return $query->where('status', Publishable::STATUS_UNPUBLISHED);
    }

    /**
     * Scope a query to only include draft content.
     *
     * @param \Illuminate\Database\Eloquent\Builder $query
     * @return \Illuminate\Database\Eloquent\Builder
     */
    public function scopeWhereDraftOrInReview(Builder $query)
    {
        return $query->where(function ($query) {
            $query->where('status', '=', Publishable::STATUS_DRAFT);
            $query->orWhere('status', '=', Publishable::STATUS_REVIEW);
        });
    }

    /**
     * Scope a query to only include draft content.
     *
     * @param \Illuminate\Database\Eloquent\Builder $query
     * @return \Illuminate\Database\Eloquent\Builder
     */
    public function scopeWhereDraft(Builder $query)
    {
        return $query->where('status', '=', Publishable::STATUS_DRAFT);
    }

    /**
     * Scope a query to only include in review content.
     *
     * @param \Illuminate\Database\Eloquent\Builder $query
     * @return \Illuminate\Database\Eloquent\Builder
     */
    public function scopeWhereInReview(Builder $query)
    {
        return $query->where('status', '=', Publishable::STATUS_REVIEW);
    }

    /**
     * Return if the current content has a status of published.
     *
     * @return boolean
     */
    public function isPublished()
    {
        return $this->status === Publishable::STATUS_PUBLISHED;
    }

    /**
     * Return if the current content has a status of un-published.
     *
     * @return boolean
     */
    public function isUnPublished()
    {
        return $this->status === Publishable::STATUS_UNPUBLISHED;
    }

    /**
     * Return if the current content is a draft copy.
     *
     * @return boolean
     */
    public function isDraft()
    {
        return $this->status === Publishable::STATUS_DRAFT;
    }

    /**
     * Return if the current content is marked as in review.
     *
     * @return boolean
     */
    public function isInReview()
    {
        return $this->status === Publishable::STATUS_REVIEW;
    }

    /**
     * Determine if viewing draft content.
     *
     * @return boolean
     */
    public function isDraftOrInReview()
    {
        return $this->isDraft() || $this->isInReview();
    }

    /**
     * Return if a published version of the current content exists.
     *
     * @return boolean
     */
    public function hasPublishedVersion()
    {
        return !empty($this->getPublishedVersion());
    }

    /**
     * Return if a draft|awaiting version of the current content exists.
     *
     * @return boolean
     */
    public function hasDraftOrReviewVersion()
    {
        return !empty($this->getDraftOrReviewVersion());
    }

    /**
     * Return if a unpublished version of the current content exists.
     *
     * @return boolean
     */
    public function hasUnPublishedVersion()
    {
        return !empty($this->getUnpublishedVersion());
    }

    /**
     * Return if the current content has been saved to the DB so can be published.
     *
     * @return boolean
     */
    public function canBePublished()
    {
        return $this->exists;
    }

    /**
     * Return if the current content is can be reviewed.
     * Only draft and in review status can be reviewed.
     *
     * @return boolean
     */
    public function canBeReviewed()
    {
        return $this->isDraftOrInReview();
    }

    /**
     * Return if the content can be reset. This means it
     * is currently draft and has a published
     * version to fall back to
     *
     * @return boolean
     */
    public function canBeReset()
    {
        return $this->hasDraftOrReviewVersion() && $this->hasPublishedVersion();
    }

    /**
     * Publish the content if it is a draft.
     * Un-publish the content if it is published (set to Draft)
     *
     * @return void
     */
    public function updatePublishStatus()
    {
        $this->isPublished() ? $this->unPublish() : $this->publish();
    }

    /**
     * Update the content to be in-review if it's a draft.
     * Update to draft it is already in-review
     *
     * @return void
     */
    public function updateReviewStatus()
    {
        $this->status = $this->isInReview() ? Publishable::STATUS_DRAFT : Publishable::STATUS_REVIEW;
        $this->save();
    }

    /**
     * Publish the draft content.
     * If their is already a published version this is deleted.
     *
     * @return void
     */
    private function publish()
    {
        if ($this->isDraft()) {
            $this->removeExistingVersion();
        }

        $this->status = Publishable::STATUS_PUBLISHED;
        $this->save();
        event(new ContentWasPublished($this));
    }

    /**
     * Set the published content to draft.
     *
     * @return void
     */
    private function unPublish()
    {
        $this->status = Publishable::STATUS_UNPUBLISHED;
        $this->save();
    }

    /**
     * Remove An existing version
     *
     * @return void
     */
    private function removeExistingVersion()
    {
        $published = $this->getPublishedVersion();
        if ($published) {
            $published->delete();
        }

        $unPublished = $this->getUnpublishedVersion();
        if ($unPublished) {
            $unPublished->delete();
        }
    }

    /**
     * Remove the draft version of the content so there is only the published version left.
     *
     * @return void
     * @throws CouldNotResetDraft Thrown if there was no draft version or the draft version is the only version.
     */
    public function resetDraft()
    {
        if (!$this->canBeReset()) {
            throw new CouldNotResetDraft();
        }

        $draft = $this->getDraftOrReviewVersion();
        $draft->delete();
    }

    /**
     * Return the latest version of the content using the given Id.
     * If their is not a draft version the published version will be returned.
     * If there is then that draft will be returned.
     *
     * @param integer $contentId
     * @return \Drp\SimpleCms\Content
     */
    public static function getLatestVersionById($contentId)
    {
        return self::where('content_id', $contentId)
            ->orderBy('status')
            ->firstOrFail();
    }

    /**
     * Return all the content. Only include the
     * Published content or content that only have a draft
     * i.e ones that have never been published.
     *
     * @param array $with Optional relationships to eager load
     * @return mixed
     */
    public static function getAllVersionsPrioritisedByStatus(array $with = [])
    {
        return self::with($with)
            ->get()
            ->groupBy('content_id')
            ->map(function ($content) {
                $keyed = $content->keyBy('status');
                //Order by published status
                return $keyed->get(
                    Publishable::STATUS_PUBLISHED,
                    $keyed->get(
                        Publishable::STATUS_UNPUBLISHED,
                        $keyed->get(
                            Publishable::STATUS_REVIEW,
                            $keyed->get(Publishable::STATUS_DRAFT)
                        )
                    )
                );
            })
            ->flatten();
    }

    /**
     * Return the options for the select box
     * using getAllPublishedOrFirstDraft.
     *
     * @return mixed
     */
    public static function optionsForSelect()
    {
        return self::getAllVersionsPrioritisedByStatus()
            ->pluck('title', 'content_id')
            ->toArray();
    }

    /**
     * Return a closure that is called before a study is saved.
     *
     * @return \Closure|null
     * @SuppressWarnings("unused")
     */
    public function beforeSave()
    {
        return null;
    }

    /**
     * Return a closure that is called after a study is saved.
     *
     * @return \Closure|null
     * @SuppressWarnings("unused")
     */
    public function afterSave()
    {
        return null;
    }

    /**
     * Check if the content has changed from the DB version when the attributes are applied
     *
     * @param mixed $attributes
     * @return mixed
     */
    public function isContentDirty($attributes)
    {
        if (!$this->exists()) {
            return true;
        }

        return $this->fresh()->fill($attributes)->isDirty();
    }
}
